import pandas as pd
import numpy as np
from typing import Union, Tuple
from adlinear import nmfmodel as nmf
from adlinear import ntfmodel as ntf
import math


class NMFClient:

    def __init__(self, name: str):
        self._name = name
        self._data = pd.DataFrame()
        self._model: nmf.NmfModel = nmf.NmfModel(self._data, ncomp=2)
        self._latest_h = pd.DataFrame
        self._latest_w = pd.DataFrame
        self._latest_error: float = 1.0
        self._latest_error_delta: float = 0.0
        self._niter = 0
        pass

    def reset(self):
        self._latest_error = 1.0
        self._latest_error_delta = 0.0
        self._niter = 0

    def get_ncomp(self) -> int:
        return self._model.ncomp

    def get_name(self) -> str:
        return self._name

    def set_name(self, name: str):
        self._name = name

    def set_data(self, data: pd.DataFrame):
        self._data = data
        ncomp = self._model.ncomp
        self._model = nmf.NmfModel(self._data, ncomp=ncomp)

    def set_ncomp(self, ncomp: int):
        self._model.set_ncomp(ncomp)

    def update_h(self, h_estd: pd.DataFrame):
        self._model.update_h(h_estd)
        self._latest_h = self._model.get_h()
        self._latest_w = self._model.get_w()
        err = self._latest_error
        self._latest_error = self._model.get_precision(relative=True)
        self._latest_error_delta = self._latest_error - err
        self._niter += 1
        pass

    def get_latest_h(self) -> pd.DataFrame:
        return self._latest_h

    def get_latest_error(self) -> float:
        return self._latest_error

    def get_latest_error_delta(self) -> float:
        return self._latest_error_delta

    def get_count(self) -> int:
        return self._niter


class NMFCentralizer:

    def __init__(self, nfeat: int, ):
        self._current_h = pd.DataFrame(index=range(nfeat))
        self._nmfcomp = 1
        self._nfeat = nfeat
        self._learning_rate = 0.1
        self._err = 1
        self._niter = 0
        self._perf_log = pd.DataFrame(index=[], columns=self.get_logcolumns())
        return

    @staticmethod
    def get_logcolumns():
        return ["Error before", "Last client", "Client error", "Client error delta", "Error after", "Error delta"]

    def get_perf_log(self):
        return self._perf_log

    def reset(self):
        self._current_h = pd.DataFrame(index=range(self._nfeat),
                                       columns=range(self._nmfcomp),
                                       data=np.ones((self._nfeat, self._nmfcomp)))
        self._err = 1
        self._niter = 0
        return

    def set_ncomp(self, ncomp: int):
        self._nmfcomp = ncomp
        self._current_h = pd.DataFrame(index=range(self._nfeat),
                                       columns=range(self._nmfcomp),
                                       data=np.ones((self._nfeat, self._nmfcomp)))
        return

    def set_features(self, features: Tuple[str]):
        self._nfeat = len(features)
        self._current_h = pd.DataFrame(index=features,
                                       columns=range(self._nmfcomp),
                                       data=np.ones((self._nfeat, self._nmfcomp)))
        return

    def err(self):
        return self._err

    def set_h_guess(self, h: pd.DataFrame):
        self._current_h = h.copy()

    def request_for_update(self, client: NMFClient):
        client.update_h(self._current_h)
        err_before = self._err
        h = client.get_latest_h()
        err_clt = client.get_latest_error()
        w_clt = math.exp(- self._learning_rate * err_clt)
        w_self = math.exp(- self._learning_rate * self._err) if self._niter > 0 else 0
        assert len(self._current_h.columns) == len(h.columns)
        self._current_h.columns = h.columns
        self._current_h = w_clt * h + w_self * self._current_h
        self._current_h = self._current_h / (w_clt + w_self)
        self._err = (w_clt * err_clt + w_self * err_before) / (w_clt + w_self)

        log_line = pd.DataFrame(index=[self._niter], columns=self.get_logcolumns())
        log_line.loc[self._niter, "Last client"] = client.get_name()
        log_line.loc[self._niter, "Err before"] = err_before
        log_line.loc[self._niter, "Client error"] = err_clt
        log_line.loc[self._niter, "Client error delta"] = client.get_latest_error_delta()
        log_line.loc[self._niter, "Err before"] = err_before
        log_line.loc[self._niter, "Err delta"] = self._err - err_before
        self._perf_log = pd.concat([self._perf_log, log_line], axis=0)
        self._niter += 1
        pass


class FederatedNMFConfig:

    def __init__(self,
                 nmfcentral: NMFCentralizer,
                 clients: Tuple[NMFClient]):

        self._nmfcentral: NMFCentralizer = nmfcentral
        self._clients: Tuple[NMFClient] = []
        self.set_clients(clients, [])
        pass

    def set_clients(self, clients: Tuple[NMFClient], names: Tuple[str] = []):
        nclients = len(clients)
        if len(names) == 0:
            names = [f"Client {i+1}" for i in range(nclients)]
        else:
            pass
        assert len(names) == nclients
        self._clients = clients
        for iclt, clt in enumerate(self._clients):
            clt.set_name(names[iclt])

    def set_client_names(self, names: Tuple[str] = []):
        nclients = len(self._clients)
        if len(names) == 0:
            names = [f"Client {i+1}" for i in range(nclients)]
        else:
            pass
        assert len(names) == nclients
        for iclt, clt in enumerate(self._clients):
            clt.set_name(names[iclt])

    def get_central(self) -> NMFCentralizer:
        return self._nmfcentral

    def set_central(self, central: NMFCentralizer):
        self._nmfcentral = central

    def get_clients(self):
        return self._clients

    def set_ncomp(self, ncomp: int):
        self._nmfcentral.set_ncomp(ncomp)
        for clt in self._clients:
            clt.set_ncomp(ncomp)

    def set_features(self, features: Tuple[str]):
        self._nmfcentral.set_features(features)

    def request_update_step(self, client_idx: int):
        if 0 <= client_idx <= len(self._clients):
            clt = self._clients[client_idx]
            self._nmfcentral.request_for_update(clt)

    def request_full_round(self):
        for iclt, clt in enumerate(self._clients):
            self._nmfcentral.request_for_update(clt)
            clt_count = clt.get_count()
            print(f"After checking {clt.get_name()}, trial {clt_count}: erreur: {self._nmfcentral.err()}")

    def reset_counters(self):
        self._nmfcentral.reset()
        for clt in self._clients:
            clt.reset()

    def get_perf_log(self):
        return self._nmfcentral.get_perf_log()





import functools
from transparentpath import Path
import pandas as pd
# import numpy as np
import numpy.random
import dotenv
import argparse


# from dask.distributed import Client
# from dask import distributed as dst
#

# from transparentpath import Path
# import dask.dataframe as dd
#
# client = Client("tcp://127.0.0.1:33869")
#
# dfs_path = Path("/home/cgeissler/Documents/Tests")
# ddf = dd.read_parquet(dfs_path)
#
# ddf.loc["2020-01-01":"2020-01-03"].groupby('feature_name').mean().compute()
#
#
# # In[4]:
#
#
# ind = (((ddf.groupby('feature_name').count().sum(axis=1) / (ddf.index.nunique() * ddf.shape[1])) > 0.5)).compute()
# ind.loc[ind].index.to_list()
# ddf = ddf.loc[ddf['feature_name'].isin(ind.loc[ind].index.to_list())].compute()
#
#
# # In[5]:
#
#
# a = (ddf.groupby('feature_name').max().max(axis=1)) / 2 + (ddf.groupby('feature_name').min().min(axis=1)) / 2
# b = (ddf.groupby('feature_name').max().max(axis=1)) / 2 - (ddf.groupby('feature_name').min().min(axis=1)) / 2
# ind = ddf.index
# ddf = ddf.set_index('feature_name')
# for l in a.index:
#     ddf.loc[l] = (ddf.loc[l] - a[l]) / b[l]
# ddf = ddf.reset_index()
# ddf.index = ind

Path.set_global_fs("gcs", bucket="nmf_experiments_dev", token="../cred_gcs.json")
dotenv.load_dotenv()


def collect_and_concat_files(collect_path: str, afilter: str = "", file_ext: str = ".csv") -> pd.DataFrame:
    """
    collects files in a given (transparent) path and concatenates these files in a single dataframe.
    Parameters
    ----------
    collect_path:
        path to collect the files from
    afilter:
        string (not regex).
        If not empty, only files containing the pattern as a substring are collected.
    file_ext:
        An optional extension the collected files must match with.
    Returns
    -------
        The concatenation of all collected files in a single dataframe.
    """
    collect_path = Path(collect_path)
    file_list = collect_path.glob(f"*{file_ext}")
    df_concat = pd.concat([afile.read(index_col=0) for afile in file_list
                           if afilter == "" or str(afile).find(afilter) >= 0], axis='index')
    return df_concat


def collect_features_mean_values(data_path: Path, **kwargs) -> pd.DataFrame:
    # exclusion_list=None, period="", mode: str = "mean")
    """
    collect_features_mean_values: collects files into a single matrix
    Each file represent the time history of a given feature.
    The structure of the file must comply with:
    An increasing series of dates as index;
    Object names as columns

    :Parameters
    ----------

    data_path: a directory path indicating the location of data.
        All files must be located in data_path
    exclusion_list: list of strings to avoid in data_path

    period: time interval used to group the observations.
    If unspecified or '', the whole index will be kept for each feature.
    Otherwise, a time interval.

    mode: 'mean' (default), 'first' or 'last' to determine which value to extract

    :returns:
    a DataFrame containing the features (attributes) as columns, the observations (e.g stocks) as lines
    """

    # total files list
    exclusion_list = kwargs.get("exclusion_list", None)
    measurement_dates = kwargs.get("dates", None)
    mode = kwargs.get("mode", "mean")
    do_save = kwargs.get("save", True)
    out_path = kwargs.get("out_path", "")
    if exclusion_list is None:
        exclusion_list = []
    flist = [
        x
        for x in data_path.glob("*")
        if functools.reduce(lambda x, y: x and y, [str(x).find(excl) < 0 for excl in exclusion_list], True)
    ]
    matrix = pd.DataFrame(index=[], columns=[])
    nmax = 0
    for fl in flist:
        feature = fl.stem
        hdf = pd.read_csv(data_path / fl, index_col=0, parse_dates=True).fillna(method="ffill")
        if measurement_dates != "" and measurement_dates is not None:
            hdf = hdf.loc[measurement_dates, :]
            str_dates = f"Set_{len(hdf.index)}_dates"
        else:
            str_dates = ""
        if mode == "mean":
            hdf = pd.DataFrame(hdf.mean(axis=0))
        elif mode == "first":  # last
            hdf = pd.DataFrame(hdf.iloc[0])
        elif mode == "last":
            hdf = pd.DataFrame(hdf.iloc[-1])
        elif mode == "random":
            irnd = numpy.random.Generator.choice(range(len(hdf.index)), 1)[0]
            hdf = pd.DataFrame(hdf.iloc[irnd])
        hdf.columns = [feature]
        if nmax > 0 and len(hdf) >= nmax + 50:
            print((len(hdf), "stocks before."))
            hdf = hdf.loc[matrix.index, :]
            print((len(hdf), "stocks after."))

        if len(hdf) > nmax:
            nmax = len(hdf)
        matrix = pd.concat([matrix, hdf], axis=1, sort=False)
        if do_save:
            mat_name = f"Features_from_{data_path}_{mode}"
            if str_dates != "":
                mat_name += f"at_{str_dates}__.csv"
            matrix.to_csv(Path(out_path) / mat_name)

    return matrix.dropna(how="all")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Files collector",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-p", "--path", type=str, default="", help="Files localization")
    parser.add_argument("-f", "--filter", type=str, default="", help="")
    parser.add_argument("-o", "--output", type=str, default="output.csv", help="Output localization")

    args = parser.parse_args()
    filepath = Path(args.path)
    file_filter = args.filter
    output = Path(args.output, fs="local")

    df_outputs = collect_and_concat_files(collect_path=filepath,
                                          afilter=file_filter)
    output.write(df_outputs)
    pass



import pandas as pd

from adlinear import fed_nmtf as fd
from adlinear import utilities as utl
import root_path
from transparentpath import Path
import os
import dotenv
from adlinear import nmfmodel as nmf

dotenv.load_dotenv()

if __name__ == "__main__":

    # Data localization
    rootpath = root_path.get_root_path()

    rd_path = Path(rootpath / os.getenv("rd_subpath"), fs="local")
    federated_path: object = Path(rd_path / os.getenv("federated_subpath"), fs="local")

    data_path: object = Path(federated_path / os.getenv("data_subpath"), fs="local")

    temp_data_path = Path(data_path / "temp_data/", fs="local")
    pictures_path = Path(data_path / "Pics", fs="local")
    out_data_path = Path(federated_path / "results/", fs="local")

    res_path = Path(federated_path / os.getenv("results_subpath"), fs="local")

    if os.getenv("fed_do_epithor", "False").lower() == "true":

        epithor_filename: str = os.getenv("fed_epithor_filename")
        df_epithor: pd.DataFrame = Path(data_path / epithor_filename, fs="local").read()
        df_epithor.drop(columns=df_epithor.columns[0:4], inplace=True)
        df_epithor.drop(labels=['Code', 'Suivi', 'Annee'], axis='columns', inplace=True)
        df_dense, _ = utl.drop_sparse_columns(df_epithor, minimum_fill_rate=0.8)
        df_epi_normalized, df_epi_means, df_epistdevs = utl.normalize_by_columns(df_dense, make_positive=True)
        df_n_lines = len(df_dense.index)
        epithor_nb_feat = len(df_dense.columns)
        epithor_ncomp = 10

        # initialisation
        global_model = nmf.NmfModel(df_epi_normalized, epithor_ncomp)
        global_h: pd.DataFrame = global_model.get_h()
        print(f"Erreur globale de la factorisation a {epithor_ncomp} composants: {global_model.get_precision()}")
        # centralizer
        centralizer: fd.NMFCentralizer = fd.NMFCentralizer(nfeat=epithor_nb_feat)
        centralizer.set_ncomp(ncomp=epithor_ncomp)
        # clients
        nb_clients: int = int(os.getenv("fed_nb_clients", 5))
        df_slice: int = int(df_n_lines / nb_clients)
        # création des clients
        clients = [fd.NMFClient(name="") for _ in range(nb_clients)]
        clt_names = []
        # attribution des données aux clients
        for icl, client in enumerate(clients):
            client.set_data(df_epi_normalized.iloc[int(icl * df_slice): int((icl+1) * df_slice), :])
        # config
        fed_config: fd.FederatedNMFConfig = fd.FederatedNMFConfig(nmfcentral=centralizer,
                                                                  clients=clients)
        fed_config.set_client_names([])

        fed_config.set_ncomp(epithor_ncomp)
        fed_config.set_features(df_epi_normalized.columns)
        # for _ in range(10):
        #     fed_config.request_full_round()

        fed_config.reset_counters()
        fed_config.get_central().set_h_guess(global_h)
        for _ in range(5):
            fed_config.request_full_round()
        fed_config.get_perf_log().loc["Error before"].plot()
        pass


import math
from . import nmfmodel
import seaborn as sns
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
# from .utilities import *
from . import utilities as utl
from typing import Union, Tuple
from adinference import admlp


class NCompEstimator:

    def __init__(self, inference_model: admlp.AdMlp):
        self._inference_model = inference_model
        pass

    def get_predictions(self, data_for_inference: np.ndarray):
        normalizers = self._inference_model.get_normalization()
        add_csts = pd.DataFrame(normalizers.iloc[0, :])
        mult_csts = pd.DataFrame(normalizers.iloc[1, :])
        norm_data_for_inference, _, _ = utl.normalize_by_columns(data_for_inference,
                                                                 add_cst=add_csts, mult_cst=mult_csts)
        preds = self._inference_model.predict(norm_data_for_inference)
        return preds

    def estimate_ncomp(self, mat: Union[pd.DataFrame, np.ndarray] = [],
                       df_mini_screeplots=None, ncmin: int = 2,
                       ncmax: int = 35, voting: str = "max"):

        if df_mini_screeplots is None:
            df_screeplot = nmfmodel.generate_scree_plot(mat, ncmin=ncmin, ncmax=ncmax)
            df_mini_screeplots = pd.DataFrame(index=[], columns=[])
            df_mini_screeplots = nmfmodel.collect_windows_from_scree_plot(df_collected_windows=df_mini_screeplots,
                                                                          df_scree_plot=df_screeplot,
                                                                          window_size=6)
        data_for_inference = df_mini_screeplots.drop([col for col in df_mini_screeplots.columns if
                                                      col.find("_ncomp") >= 0 or col.find("_entropy") >= 0
                                                      or col.find("Position") >= 0],
                                                     axis="columns", inplace=False).to_numpy(dtype=np.float64)

        preds = pd.DataFrame(index=df_mini_screeplots.index, data=self.get_predictions(data_for_inference))
        synthegram = pd.DataFrame(index=df_mini_screeplots.index, columns=["Score"], data=0)
        for irow, idx in enumerate(df_mini_screeplots.index):
            tried_ncomp = df_mini_screeplots.loc[idx, "Tried_ncomp_0"]
            shifts = [np.argmax(preds.iloc[irow, :])] if voting == "max" else range(preds.shape[1])
            for shift in shifts:
                pred = preds.iloc[irow, shift]
                if shift == 0:
                    synthegram.loc[:tried_ncomp] += pred
                elif shift == 5:
                    synthegram.loc[tried_ncomp + 5:] += pred
                    pass
                else:
                    if tried_ncomp + shift <= synthegram.index[-1]:
                        synthegram.loc[tried_ncomp + shift] += 5 * pred
                    pass
        return synthegram, preds, df_mini_screeplots

    pass



import numpy as np
import pandas as pd
import adinference
from adinference import admlp
from adlinear import utilities as utl
from adlinear import nmf_k_estimator as ke
from adlinear import nmfmodel as nmf
import root_path
from transparentpath import Path
import os
import dotenv
from randomgenerators import randomgenerators as rng


dotenv.load_dotenv()

if __name__ == "__main__":

    # Data localization
    rootpath = root_path.get_root_path()

    rd_path = Path(rootpath, fs="local") / os.getenv("rd_subpath")
    optimal_k_path: Path = Path(rd_path / os.getenv("optimal_k_subpath"), fs="local")

    data_path: Path = Path(optimal_k_path, fs="local") / os.getenv("data_subpath")
    models_path: Path = Path(optimal_k_path, fs="local") / os.getenv("models_subpath")

    temp_data_path = Path(data_path, fs="local") / "temp_data/"
    pictures_path = Path(data_path, fs="local") / "plot"
    out_data_path = Path(optimal_k_path, fs="local") / "results/"
    optk_model_dir = os.getenv("optk_model_dir", "model_MLP_p82")

    res_path = Path(optimal_k_path, fs="local") / os.getenv("results_subpath")

    raw_screeplots_filename: str = os.getenv("raw_screeplot_file")

    df_raw_screeplots: pd.DataFrame = (Path(data_path, fs="local") / raw_screeplots_filename).read(index_col=0)
    df_train = df_raw_screeplots.drop([col for col in df_raw_screeplots.columns if col.find("comp") >= 0],
                                      axis='columns', inplace=False)
    df_train = df_train.drop([col for col in df_train.columns if col.find("entropy") >= 0],
                             axis='columns', inplace=False)
    train_file_path = Path(data_path, fs="local") / "screeplots.csv"
    train_file_already_saved = True
    if not train_file_already_saved:
        train_file_path.write(df_train)

    mymlp = adinference.admlp.AdMlp(models_path, "model_MLP_p82")
    mymlp.load_model()

    my_k_estimator = ke.NCompEstimator(inference_model=mymlp)

    est_do_random_matrix = os.getenv("est_do_random_matrix", "false")
    if est_do_random_matrix.lower() == "true":
        nb_clusters = 10
        h_size = 50
        h_avg = int(h_size / nb_clusters)
        w_size = 1000
        w_avg = int(w_size / nb_clusters)
        icorrmin = 0.80
        icorrmax = 0.99
        xcorrmin = 0.01
        xcorrmax = 0.30

        mat, _, _, _ = rng.generate_nmf_reconstruction(n_comp=nb_clusters, n_feat=h_size, n_obs=w_size,
                                                       avg_w_clust=w_avg, avg_h_clust=h_avg,
                                                       h_icorr_min=icorrmin, h_xcorr_max=xcorrmax,
                                                       w_icorr_min=icorrmin, w_xcorr_max=xcorrmax)
        preds = my_k_estimator.estimate_ncomp(mat=mat)

    est_do_train_data = False
    if est_do_train_data:
        train_data = df_raw_screeplots.drop(['Position'], axis="columns", inplace=False)
        # norm_train_data, _, _ = utl.normalize_by_columns(train_data, make_positive=False)
        norm_train_data = train_data.copy()
        norm_train_data.loc[:, "Tried_ncomp_0"] = train_data.loc[:, "Tried_ncomp_0"]
        norm_train_data = norm_train_data.iloc[0:16]
        ncmin = np.min(norm_train_data.index)
        ncmax = np.max(norm_train_data.index)-1
        train_synthegram = my_k_estimator.estimate_ncomp(df_mini_screeplots=norm_train_data, ncmin=ncmin, ncmax=ncmax)

    est_do_swimmer = True
    if est_do_swimmer:
        swimmer_file = "swimmer.xlsx"
        swimmer_data = pd.read_excel(str(data_path.path) + "/" + swimmer_file)
        swimmer_data = swimmer_data.iloc[:, 2:]
        # swimmer_screeplot = nmf.generate_scree_plot(swimmer_data, ncmin=2, ncmax=30)
        recalc_swimmer = False
        ncmin = 2
        ncmax = 30
        if recalc_swimmer:
            sw_synthegram, sw_preds, sw_mini_screeplots = my_k_estimator.estimate_ncomp(mat=swimmer_data,
                                                                                        ncmin=ncmin, ncmax=ncmax)
            pass
        else:
            sw_mini_screeplots = pd.read_csv(str(res_path) + "/swimmer_mini_screeplots.csv", index_col=0)
            sw_synthegram, sw_preds, _ = \
                my_k_estimator.estimate_ncomp(df_mini_screeplots=sw_mini_screeplots,
                                              ncmin=ncmin, ncmax=ncmax)
        sw_synthegram.to_csv(str(res_path) + "/swimmer_synthegram.csv")
        sw_preds.to_csv(str(res_path) + "/swimmer_preds.csv")
        sw_mini_screeplots.to_csv(str(res_path) + "/swimmer_mini_screeplots.csv")

    est_do_sausage = True
    if est_do_sausage:
        sausage_file = "Sausage Raw NIR.xlsx"
        sausage_data = pd.read_excel(str(data_path.path) + "/" + sausage_file)
        sausage_data = sausage_data.iloc[:, 1:]
        # sausage_screeplot = nmf.generate_scree_plot(sausage_data, ncmin=2, ncmax=30)
        recalc_sausage = True
        ncmin = 2
        ncmax = 20
        if recalc_sausage:
            sa_synthegram, sa_preds, sa_mini_screeplots = my_k_estimator.estimate_ncomp(mat=sausage_data,
                                                                                        ncmin=ncmin, ncmax=ncmax)
            pass
        else:
            sa_mini_screeplots = pd.read_csv(str(res_path) + "/sausage_mini_screeplots.csv", index_col=0)
            sa_synthegram, sa_preds, _ = \
                my_k_estimator.estimate_ncomp(df_mini_screeplots=sa_mini_screeplots,
                                              ncmin=ncmin, ncmax=ncmax)
        sa_synthegram.to_csv(str(res_path) + "/sausage_synthegram.csv")
        sa_preds.to_csv(str(res_path) + "/sausage_preds.csv")
        sa_mini_screeplots.to_csv(str(res_path) + "/sausage_mini_screeplots.csv")

        pass





from pathlib import Path
from transparentpath import TransparentPath

import kfp.dsl as dsl
import kfp
from kubernetes.client.models import V1Toleration
import argparse


class Output:
    def __init__(self, dirpath: TransparentPath, name: str, suf: str = "", prefix: str = ""):

        if dirpath is None:
            raise ValueError("root directory for outputs can not be None")

        self.dirpath = dirpath
        self.path = None
        attempts = 0
        if prefix != "":
            name = f"{name}_{prefix}"

        if suf != "":
            path = (dirpath / f"{name}_{attempts}").with_suffix(suf)
            while path.is_file():
                if attempts > 1e9:
                    raise ValueError(f"Could not create output file after {attempts} attempts")
                attempts += 1
                path = (dirpath / f"{name}_{attempts}").with_suffix(suf)

            self.path = path
            self.path.touch()
            print(f"Making file {self.path}")
        else:
            path = dirpath / f"{name}_{attempts}"
            while path.is_dir():
                if attempts > 1e9:
                    raise ValueError(f"Could not create output dir after {attempts} attempts")
                attempts += 1
                path = dirpath / f"{name}_{attempts}"

            self.path = path
            (self.path / "locker").touch()
            print(f"Making directory {self.path}")


# noinspection PyUnresolvedReferences
def function(prog, prog_name, image, memory_limit, cpu, items, tag, argnames, preemptible, timeout):
    if "/" not in image:
        image = f"eu.gcr.io/crested-acumen-297311/{image}:{tag}"
    else:
        image = f"{image}:{tag}"

    arguments = [prog]
    for argname in argnames:
        arguments.append(f"--{argname}")
        arguments.append(getattr(items, argname))

    container_op = dsl.ContainerOp(
        name=prog_name,
        image=image,
        command=["python"],
        arguments=arguments,
    )
    # Single job specification
    container_op.add_resource_limit("memory", memory_limit)
    container_op.add_resource_limit("cpu", cpu)
    container_op.add_resource_request("memory", memory_limit)
    container_op.add_resource_request("cpu", cpu)
    if timeout is not None:
        container_op.set_timeout(timeout)
    if preemptible:
        container_op.add_toleration(
            V1Toleration(key="dedicated", operator="Equal", value="standard-preemptible", effect="NoSchedule")
        )
    return container_op


def make_yaml(
    program,
    iterations,
    iterfile,
    image_name,
    parallelism,
    ram,
    cpus,
    suffix,
    root,
    no_out,
    additionnal_args,
    prefix,
    tag,
    preemptible,
    timeout,
):

    if program is None:
        raise ValueError("Program path in docker can not be None")

    program_path = Path(program)
    additionnal_args = {item.split(":")[0]: item.split(":")[1] for item in additionnal_args.split(" ")}
    if iterfile != "":
        iterfile = TransparentPath(iterfile, fs="local")
        if not iterfile.is_file():
            raise FileNotFoundError(f"Could not find iteration file '{iterfile}'")

    if ram != "" and not ram.endswith("G") and not ram.endswith("M"):
        raise ValueError("Malformed RAM argument. Should end by 'G' or 'M'")

    if not isinstance(iterations, int) and iterfile == "":
        raise TypeError("Number of iterations must be an integer")

    if iterations < 0 and iterfile == "":
        raise ValueError("Number of iterations must be greater than 0")

    if image_name == "":
        image_name = program_path.stem

    @dsl.pipeline(name="function")
    def pipeline():
        if iterfile != "":
            iterable = iterfile.read()
        else:
            iterable = [{} for _ in range(iterations)]
        if not no_out:
            TransparentPath.set_global_fs("gcs")  # Needs GOOGLE_APPLICATION_CREDENTIALS
        for item in iterable:
            if not no_out:
                item["output"] = str(Output(TransparentPath(root), program_path.stem, suffix, prefix).path)
            for arg in additionnal_args:
                item[arg] = additionnal_args[arg]
            if "name" in item:
                raise ValueError("The argument 'name' is reserved by KubeFlow, you can not use it asone of your "
                                 "program's arguments")

        argnames = []
        for item in iterable:
            argnames += list(item.keys())
        argnames = set(argnames)
        for item in iterable:
            for argname in argnames:
                if argname not in item:
                    raise ValueError(f"Argument '{argname}' is not present in all your runs arguments.")

        with dsl.ParallelFor(iterable, parallelism=parallelism) as item:  # On max 10 machines at a time
            function(
                str(program_path), program_path.stem, image_name, ram, cpus, item, tag, argnames, preemptible, timeout
            )

    yaml_file = program_path.with_suffix(".yaml").name
    # noinspection PyUnresolvedReferences
    kfp.compiler.Compiler().compile(pipeline, yaml_file)
    print(f"Created {yaml_file} file")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="A tool to launch one python program several times in parallel on GCP cluster. The program must "
                    "be in a docker image.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-p", "--program", type=str, default=None, help="The relative path to the program inside the docker image"
    )
    parser.add_argument("-i", "--iterations", type=int, default=1, help="The number of times the program must be ran")
    parser.add_argument(
        "-I",
        "--iterfile",
        type=str,
        default="",
        help="Replaces -i, path to a .json file where each line is a set of arguments to pass to the program. The "
             "program will then be launched once per line. See tests/iterfile.json for an example.",
    )
    parser.add_argument(
        "-m", "--max_executions", type=int, default=15, help="Max number of runs that can exist at a time"
    )
    parser.add_argument(
        "-d",
        "--docker",
        type=str,
        default="",
        help="Docker image to use. Default will be eu.gcr.io/crested-acumen-297311/<program stem>. If only the name "
             "if specified, the image will be eu.gcr.io/crested-acumen-297311/<name>. One can also specify the full "
             "image path, for example 'eu.gcr.io/second-capsule-253207/coucou'.",
    )
    parser.add_argument(
        "-a", "--args", type=str, default="", help="additionnal arguments to pass to the program to run"
    )
    parser.add_argument(
        "-n",
        "--nooutputs",
        action="store_true",
        help="If specified, the program is not supposed to write any file and no random output file will be created",
    )
    parser.add_argument(
        "-r",
        "--root",
        type=str,
        default=None,
        help="Root path where single run output should be stored. Useless if **-n** is passed.",
    )
    parser.add_argument("-R", "--ram", type=str, default="2G", help="Max ram used by one single run")
    parser.add_argument("-c", "--cpu", type=str, default="1", help="Max number of CPUs used by one single run")
    parser.add_argument(
        "-s",
        "--suffix",
        type=str,
        default="",
        help="Suffix of the random named file to store single run output. If '' (default), will create a directory "
             "instead.",
    )
    parser.add_argument("-P", "--prefix", type=str, default="", help="Prefix to put in the program output names")
    parser.add_argument("-t", "--tag", type=str, default="latest", help="Docker image tag")
    parser.add_argument(
        "-k", "--preemptible", action="store_true", help="Tells the program that it can use preemptible VMs"
    )
    parser.add_argument(
        "-T", "--timeout", type=int, default=None, help="Tells one single run to kill itself after a"
                                                        " given time (in seconds) (default = None)"
    )
    args = parser.parse_args()

    make_yaml(
        args.program,
        args.iterations,
        args.iterfile,
        args.docker,
        args.max_executions,
        args.ram,
        args.cpu,
        args.suffix,
        args.root,
        args.nooutputs,
        args.args,
        args.prefix,
        args.tag,
        args.preemptible,
        args.timeout,
    )



# import itertools
# import os
import argparse
from transparentpath import Path
import numpy as np
import pandas as pd
from adlinear import nmfmodel as nmf
from randomgenerators import randomgenerators as rng
import dotenv

Path.set_global_fs("gcs", bucket="nmf_experiments_dev", token="cred_gcs.json")
dotenv.load_dotenv()

if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Creates the yaml file to launch several iterations of a program on GCP",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("-o", "--output", type=str, default="wfold_scree_plots.csv", help="Output file path")
    parser.add_argument("-r", "--runs", type=int, default=1, help="Number of iterations per node")
    parser.add_argument("-w", "--writefreq", type=int, default=1, help="Write frequency")
    parser.add_argument("-e", "--epsilon", type=float, default=0.10, help="Epsilon")
    parser.add_argument("-m", "--ncmin", type=int, default=2, help="Minimum number of components")
    parser.add_argument("-M", "--ncmax", type=int, default=45, help="Maximum number of components")
    parser.add_argument("-W", "--w_avg", type=int, default=2, help="Average size of W clusters")
    parser.add_argument("-H", "--h_avg", type=int, default=10, help="Average size of H clusters")
    # parser.add_argument("-a", "--wclust_f", type=float, default=1.0, help="#W clusters to #H clusters maximum ratio")
    parser.add_argument("-i", "--icorrmin", type=float, default=0.5, help="Minimum intra-cluster correlation")
    parser.add_argument("-I", "--icorrmax", type=float, default=0.95, help="Maximum intra-cluster correlation")
    parser.add_argument("-x", "--xcorrmin", type=float, default=0.0, help="Minimum extra-cluster correlation")
    parser.add_argument("-X", "--xcorrmax", type=float, default=0.5, help="Maximum extra-cluster correlation")
    parser.add_argument("-p", "--do_entropy", type=bool, default=False, help="Optional entropy calculation")
    args = parser.parse_args()
    outputpath = Path(args.output)
    epsilon = args.epsilon
    ncmin = args.ncmin
    ncmax = args.ncmax
    nruns = args.runs
    # wclust_f = args.wclust_f
    w_avg = args.w_avg
    h_avg = args.h_avg
    do_ent = args.do_entropy
    icorrmin = args.icorrmin
    icorrmax = args.icorrmax
    xcorrmin = args.xcorrmin
    xcorrmax = args.xcorrmax
    writefreq = args.writefreq

    print(f"Will execute {nruns} run(s), will write output(s) in {outputpath} every {writefreq} iteration(s)")

    rand_norms = True
    df_mini_scree_plots = pd.DataFrame(index=[], columns=[])
    for itrial in range(nruns):
        nb_clusters = np.random.randint(low=ncmin, high=ncmax)
        intra_corr = np.random.uniform(low=icorrmin, high=icorrmax)
        xtra_corr = np.random.uniform(low=xcorrmin, high=xcorrmax)
        h_size = h_avg * nb_clusters
        # w_size = np.random.randint(low=max(nb_clusters, int(h_size / 5.0)), high=5*h_size)
        w_size = np.random.randint(low=max(nb_clusters, int(h_size / 5.0)), high=5 * h_size)
        eps = np.random.uniform(low=0.0, high=epsilon)
        # TODO: lift the constraint that W and H have same number of clusters
        # w_avg = h_avg
        # nb_w_clusters = max(int(w_size / w_avg), 1)
        nb_w_clusters = nb_clusters
        w_avg = int(w_size / nb_w_clusters)
        # wclust_factor = np.random.uniform(0.0, wclust_f)

        # noinspection PyTupleAssignmentBalance
        generated_M, _, _, _ = rng.generate_nmf_reconstruction(n_comp=nb_clusters, n_feat=h_size, n_obs=w_size,
                                                               h_icorr_min=icorrmin, h_xcorr_max=xcorrmax,
                                                               w_icorr_min=xcorrmin, w_xcorr_max=icorrmax,
                                                               random_norms=rand_norms, epsilon=eps, avg_w_clust=w_avg,
                                                               avg_h_clust=h_avg)
        rnstr = "RandNorms" if rand_norms else "ConstNorms"

        df_scree_plot = nmf.generate_scree_plot(generated_M, ncmin=ncmin, ncmax=ncmax, known_ncomp=nb_clusters,
                                                do_entropies=do_ent)
        df_mini_scree_plots = nmf.collect_windows_from_scree_plot(df_collected_windows=df_mini_scree_plots,
                                                                  df_scree_plot=df_scree_plot, known_ncomp=nb_clusters,
                                                                  ncomp_margin=5)

        if itrial % writefreq == 0 or itrial == nruns-1:
            print(f"Saving iteration {itrial} in {outputpath}")
            # assert not df_mini_scree_plots.empty
            outputpath.write(df_mini_scree_plots)
            # df_scree_plot.to_csv(args.output.split(".csv")[0]+f"_scp{itrial}_nc{nb_clusters}.csv")
            assert outputpath.isfile()
            assert not outputpath.read().empty
